from typing import Self
from pymongo import ReturnDocument
from .base_model import BaseModel
from .fields import StringField, IntField
from ..errors import IntegrityError
from ..context import ctx


class Counter(BaseModel):
    """
    Counter offers an easy way of achieving SQL database style sequences that MongoDB
    doesn't have built in.
    """

    COLLECTION = "counters"
    KEY_FIELD = "key"

    key = StringField(unique=True)
    counter = IntField(default=1)

    @classmethod
    async def get(cls, key: str) -> Self:
        coll = ctx.db.meta.conn[cls.collection]
        result = await coll.find_one_and_update(
            {"key": key},
            {"$inc": {"counter": 1}},
            upsert=True,
            return_document=ReturnDocument.AFTER
        )
        return cls(result)

    @classmethod
    async def get_counter(cls, key: str) -> int:
        cnt = await cls.get(key)
        assert cnt.counter is not None
        return cnt.counter

    async def next(self) -> "Counter":
        if not self.key:
            raise IntegrityError("counter key is not initialised")
        return await self.__class__.get(self.key)

    async def drop(self) -> None:
        coll = ctx.db.meta.conn[self.collection]
        await coll.delete_one({"_id": self.id})



import asyncio
from typing import AsyncGenerator
from motor.motor_asyncio import AsyncIOMotorCollection
from pymongo import ReturnDocument
from .base_queue import BaseQueue
from .types import TBaseTask
from ..config import MongoQueueConfig
from ..util import now
from ..context import ctx


class MongoQueue(BaseQueue):
    _cfg: MongoQueueConfig
    _initialised: bool = False

    def __init__(self, cfg: MongoQueueConfig):
        super().__init__()
        self._cfg = cfg

    async def initialise(self):
        await self.ensure_indexes()
        await self.cleanup()

    async def ensure_indexes(self):
        ctx.log.info("setting up indexes for task queue")
        coll = self.tasks_collection
        await coll.create_index("state")
        await coll.create_index("id")

    async def cleanup(self):
        ctx.log.info("cleaning up task queue")
        old_ts = now() - self._cfg.keep_done_tasks_for
        coll = self.tasks_collection
        res = await coll.delete_many(
            {
                "state": {"$in": ["error", "done"]},
                "created_at": {"$lt": old_ts},
            }
        )
        ctx.log.info("cleaning up completed, %d tasks removed", res.deleted_count)

    @property
    def tasks_collection(self) -> AsyncIOMotorCollection:
        return ctx.db.meta.conn[self._cfg.tasks_collection]

    async def enqueue(self, task: TBaseTask) -> bool:
        if not self._initialised:
            await self.initialise()
        super().enqueue(task)
        coll = self.tasks_collection
        doc = task.to_message()
        await coll.insert_one(doc)
        await coll.find_one_and_update(
            {"id": task.id, "state": "new"},
            {"$set": {"state": "published"}},
        )
        task.state = "published"
        return True

    async def set_task_done(self, task: TBaseTask, status_message: str = "ok") -> bool:
        if not self._initialised:
            await self.initialise()
        super().set_task_done(task, status_message)
        coll = self.tasks_collection
        doc = await coll.find_one_and_update(
            {"id": task.id},
            {"$set": {"state": "done", "status_message": status_message}},
            return_document=ReturnDocument.AFTER,
        )
        task.state = doc["state"]
        task.status_message = doc["status_message"]
        return True

    async def tasks(self, no_block: bool = False) -> AsyncGenerator[TBaseTask, None]:
        from .task import BaseTask

        ctx.log.info(f"entering tasks queue loop, registry={BaseTask.REGISTRY}")
        if not self._initialised:
            await self.initialise()

        coll = self.tasks_collection

        next_cleanup = now() + self._cfg.cleanup_interval

        while True:
            doc = await coll.find_one_and_update(
                {"state": "published"},
                {"$set": {"state": "picked_up"}},
                return_document=ReturnDocument.AFTER,
            )
            if doc:
                task = BaseTask.from_message(doc)
                ctx.log.debug(f"yielding task {task}")
                yield task
                continue

            if now() > next_cleanup:
                await self.cleanup()
                next_cleanup = now() + self._cfg.cleanup_interval

            if no_block:
                return

            await asyncio.sleep(0.1)



from typing import Optional, Dict, Any, TypeVar, Type
from .fields import StringField
from .storable_model import StorableModel
from ..types import KwArgs
from ..errors import IntegrityError, MissingSubmodel, WrongSubmodel, UnknownSubmodel

TStorableSubmodel = TypeVar("TStorableSubmodel", bound="StorableSubmodel")
TStorableSubmodelType = Type[TStorableSubmodel]


class StorableSubmodel(StorableModel):
    """
    - Make the base class first. Set COLLECTION explicitly - it is not
      generated automatically from the class name but inherited instead.
    - Subclass your base class to define submodels. Set SUBMODEL to a string
      that will identify your submodel in the DB.
    - You can further subclass your submodels. To avoid saving such abstract
      intermediate models do not set SUBMODEL.
    - Register the submodel with the base model

    It is possible to register an arbitrary function instead of a proper class.
    It may be particularly useful if the correct class depends on something
    other than `submodel` field. The function will get **data from the DB and
    should return an model object.

    If you decide to do it you will likely have to override _preprocess_query()
    on your submodels to keep the expected find/destroy/update behaviour

    """

    SUBMODEL: Optional[str] = None

    submodel = StringField(required=True)

    __submodel_loaders__: Dict[str, TStorableSubmodelType] = None

    def __init__(self, attrs: Optional[Dict[str, Any]] = None, **_kwargs: KwArgs) -> None:
        if attrs is None:
            attrs = {}
        super().__init__(attrs)

        if self.is_new:
            if not self.SUBMODEL:
                raise IntegrityError(f"Attempted to create an object of abstract model {self.__class__.__name__}")
            if "submodel" in attrs:
                raise IntegrityError("Attempt to override submodel for a new object")
            self.submodel = self.SUBMODEL
        else:
            if not self.submodel:
                raise MissingSubmodel(f"{self.__class__.__name__} has no submodel in the DB. A bug?")
            self._check_submodel()

    def _check_submodel(self):
        if self.submodel != self.SUBMODEL:
            raise WrongSubmodel(f"Attempted to load {self.submodel} as"
                                f" {self.__class__.__name__}. Correct submodel"
                                f" would be {self.SUBMODEL}. A bug?")

    async def validate(self):
        await super().validate()
        self._check_submodel()

    @classmethod
    def register_submodel(cls, name: str, ctor: TStorableSubmodelType) -> None:
        """
        name arg is required to provide a way of registering constructor functions
        instead of model classes. While a submodel class has a SUBMODEL property, a
        function will unlikely have it.

        :param name: submodel name
        :param ctor: class or constructor function
        :return:
        """
        if cls.SUBMODEL:
            raise IntegrityError("Attempted to register submodel with another submodel")
        if not cls.__submodel_loaders__:
            cls.__submodel_loaders__ = {}
        if name in cls.__submodel_loaders__:
            raise IntegrityError(f"Submodel {name} already registered")
        cls.__submodel_loaders__[name] = ctor

    @classmethod
    def _preprocess_query(cls, query: Dict[str, Any]) -> Dict[str, Any]:
        if cls.SUBMODEL:
            query.update({"submodel": cls.SUBMODEL})
        return query

    @classmethod
    def _ctor(cls: TStorableSubmodelType, attrs: dict[str, Any] | None, **kwargs: KwArgs) -> TStorableSubmodel:
        if "submodel" not in attrs:
            raise MissingSubmodel(f"{cls.__name__} has no submodel in the DB. Bug?")
        if not cls.__submodel_loaders__:
            return cls(attrs, **kwargs)
        submodel_name = attrs["submodel"]
        if submodel_name not in cls.__submodel_loaders__:
            raise UnknownSubmodel(f"Submodel {submodel_name} is not registered with {cls.__name__}")
        return cls.__submodel_loaders__[submodel_name](attrs, **kwargs)



from croydon import ctx
from croydon.command import Command
from app.tasks.worker import Worker


class Tasks(Command):

    NAME = "tasks"
    HELP = "run task processing worker"

    async def run_async(self) -> None:
        from app import app
        app.initialise()

        wrk = Worker()
        future = wrk.run()
        try:
            await future
        except KeyboardInterrupt:
            ctx.log.info("SIGINT received, terminating worker")



from typing import Dict, Any, Iterable, Optional, TypeVar, Type, List
from copy import deepcopy
from pydantic import BaseModel as PydanticModel
from pymongo.errors import OperationFailure
from ..context import ctx
from ..errors import DoNotSave
from ..db import Shard, ObjectsCursor
from .fields import ObjectIdField
from .meta_model import MetaModel
from ..types import KwArgs

TBaseModel = TypeVar("TBaseModel", bound="BaseModel")
TPydanticModel = TypeVar("TPydanticModel", bound=PydanticModel)
undef = object()


class BaseModel(MetaModel):

    id: ObjectIdField = ObjectIdField()
    _initial_state: dict[str, Any]

    def __init__(self, attrs: Optional[Dict[str, Any]] = None, **_kwargs: KwArgs) -> None:
        if attrs is None:
            attrs = {}

        for field, descriptor in self._fields.items():
            if field == "id":
                value = attrs.get("_id", undef)
            else:
                value = attrs.get(field, undef)
            if value is undef:
                descriptor.set_default(self)
            else:
                # setting via __dict__ will ignore descriptors
                setattr(self, field, value)
        self._set_initial_state()

    async def validate_all(self) -> None:
        for descriptor in self._fields.values():
            await descriptor.validate(self)
        await self.validate()

    async def validate(self) -> None:
        pass

    @property
    def is_new(self) -> bool:
        return self.id is None

    def _set_initial_state(self) -> None:
        self._initial_state = {
            field: deepcopy(getattr(self, field))
            for field in self._fields
        }

    def is_modified(self) -> bool:
        for field in self._fields:
            value = self.__dict__[field]
            if value != self._initial_state[field]:
                return True
        return False

    async def _before_save(self) -> None:
        pass

    async def _before_validation(self) -> None:
        pass

    async def _before_delete(self) -> None:
        pass

    async def _after_save(self, is_new: bool) -> None:
        pass

    async def _after_delete(self) -> None:
        pass

    async def _save_to_db(self) -> None:
        pass

    async def _delete_from_db(self) -> None:
        pass

    async def invalidate(self, **kwargs: Dict[str, Any]) -> None:
        pass

    async def destroy(self: TBaseModel, skip_callback: bool = False, invalidate_cache: bool = True) -> TBaseModel:
        if self.is_new:
            return self
        if not skip_callback:
            await self._before_delete()

        if invalidate_cache:
            await self.invalidate()

        await self._delete_from_db()
        if not skip_callback:
            await self._after_delete()

        self.id = None
        return self

    async def save(self: TBaseModel, skip_callback: bool = False, invalidate_cache: bool = True) -> TBaseModel:
        is_new = self.is_new

        if not skip_callback:
            try:
                await self._before_validation()
            except DoNotSave:
                return self
        await self.validate_all()

        if not skip_callback:
            try:
                await self._before_save()
            except DoNotSave:
                return self

        if invalidate_cache:
            await self.invalidate()

        await self._save_to_db()

        self._set_initial_state()
        if not skip_callback:
            await self._after_save(is_new)

        return self

    def __repr__(self) -> str:
        attributes = ["%s=%r" % (a, getattr(self, a)) for a in self._fields]
        return "%s(\n    %s\n)" % (self.__class__.__name__, ",\n    ".join(attributes))

    def __eq__(self: "BaseModel", other: "BaseModel") -> bool:
        if self.__class__ != other.__class__:
            return False
        for field in self._fields:
            if hasattr(self, field):
                if not hasattr(other, field):
                    return False
                if getattr(self, field) != getattr(other, field):
                    return False
            elif hasattr(other, field):
                return False
        return True

    def __ne__(self: "BaseModel", other: "BaseModel") -> bool:
        return not self.__eq__(other)

    async def to_dict_ext(self,
                          fields: Iterable[str] | None = None,
                          include_restricted: bool = False,
                          convert_id: bool = False) -> Dict[str, Any]:
        if fields is None:
            fields = self._fields.keys()

        result = {}
        for field_name in fields:
            descriptor = self._fields.get(field_name)
            if descriptor is not None:
                if descriptor.restricted and not include_restricted:
                    continue
                value = getattr(self, field_name)
                if field_name == "id" and convert_id:
                    field_name = "_id"
                if callable(value):
                    continue
            else:
                computed = self._computed_fields.get(field_name)
                if not computed:
                    continue
                value = getattr(self, field_name)()
                if computed.is_async:
                    value = await value
                if isinstance(value, ObjectsCursor):
                    value = [x.to_dict() for x in await value.all()]

            result[field_name] = value

        return result

    def to_dict(self, fields: Iterable[str] | None = None,
                include_restricted: bool = False,
                convert_id: bool = False) -> Dict[str, Any]:
        if fields is None:
            fields = self._fields.keys()
        field_descriptors = ((f, self._fields.get(f)) for f in fields)

        result = {}
        for field, descriptor in field_descriptors:
            if descriptor is None:
                continue
            if descriptor.restricted and not include_restricted:
                continue
            value = getattr(self, field)
            if callable(value):
                continue
            if field == "id" and convert_id:
                field = "_id"
            result[field] = value
        return result

    @classmethod
    def create(cls: Type[TBaseModel], **attrs: Any) -> TBaseModel:
        return cls(attrs)

    @classmethod
    def from_pydantic(cls: Type[TBaseModel], model: TPydanticModel) -> TBaseModel:
        return cls(model.dict(exclude_unset=True))

    @classmethod
    def _ctor(cls: Type[TBaseModel], attrs: dict[str, Any] | None, **kwargs: Any) -> TBaseModel:
        """
        This method must be passed as a constructor to ObjectsCursor instead of the class itself
        By default, it does the same that the __init__() would do, however it does additional things
        for Submodels to instantiate a different (final Submodel) class
        """
        return cls(attrs, **kwargs)

    @staticmethod
    def __get_possible_databases() -> list["Shard"]:
        return [ctx.db.meta]

    @classmethod
    async def ensure_indexes(cls, loud: bool = False, overwrite: bool = False) -> None:
        dbs = cls.__get_possible_databases()
        for index in cls._indexes:
            for _db in dbs:
                try:
                    await _db.conn[cls.collection].create_index(index.keys, **index.options)
                except OperationFailure as e:
                    details = e.details
                    if details is not None:
                        code = details.get("code")
                        code_name = details.get("codeName")
                        if code == 85 or code_name == "IndexOptionsConflict":
                            if overwrite:
                                if loud:
                                    ctx.log.debug("Dropping index %s as conflicting", index.keys)
                                await _db.conn[cls.collection].drop_index(index.keys)
                                if loud:
                                    ctx.log.debug(
                                        "Creating index with options: %s, %s", index.keys, index.options
                                    )
                                await _db.conn[cls.collection].create_index(index.keys, **index.options)
                            else:
                                ctx.log.error(
                                    "Index %s conflicts with an existing one, use overwrite param to fix it",
                                    index.keys,
                                )

    def _reload_from_model(self: TBaseModel, obj: TBaseModel) -> None:
        for field in self._fields:
            if field == "_id":
                continue
            value = getattr(obj, field)

            # setting via __dict__ will ignore descriptors
            setattr(self, field, value)

    @classmethod
    def exposed_fields(cls) -> List[str]:
        fields = cls.exposed_base_fields()
        fields.extend(cls._computed_fields.keys())
        return fields

    @classmethod
    def exposed_base_fields(cls) -> List[str]:
        return [field for field, descriptor in cls._fields.items() if not descriptor.restricted]



from typing import Optional, Any
from .abc import AbstractCache
from croydon.middleware import req_cache_ctx


class RequestLocalCache(AbstractCache):
    NAME = "RequestLocalCache"

    async def initialise(self) -> None:
        pass

    async def get(self, key: str) -> Optional[Any]:
        cache = req_cache_ctx.get()
        if cache is None:
            return None
        return cache.get(key)

    async def set(self, key: str, value: Any) -> None:
        cache = req_cache_ctx.get()
        if cache is None:
            return
        cache[key] = value
        req_cache_ctx.set(cache)

    async def has(self, key: str) -> bool:
        cache = req_cache_ctx.get()
        if cache is None:
            return False
        return key in cache

    async def delete(self, key: str) -> bool:
        had = await self.has(key)
        if had:
            cache = req_cache_ctx.get()
            del cache[key]
            req_cache_ctx.set(cache)
        return had
